
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  disply.clear()        - clear the local pixel buffer
  disply.display()      - write the buffer to the display
  disply.displayOff()   - turn the display OFF
  disply.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  disply.init()         - initialise the display
  disply.invertDisplay()- inverted display mode
  disply.end()          - free memory used by the display
  disply.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  disply.normalDisplay()- normal display mode
  disply.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  // if DispDel > 0 then we wait for DispDel cycles to complete before update
  if (DispNOP) {return;}  // this function is aborted, no I2C activity
  
  if (DispClr) {Display_clear(); DispClr = false; return;}
  if (DispDisp) {Display_display(); DispDisp = false; DispClr = true; return;}
  if (DispDel > 0) {DispDel--; return;} // don't update display if DispDel > 0

  // we refresh individual displays based on a counter, DispCnt
  // set DispCnt -ve to prevent the display update process
  DispCnt--;
  if (DispCnt < 1) {
    // DispCnt is reset for display updates, so can be different for each
    DispCnt = 1;  // default, will trigger on each cycle
    if (DispLock) {DispMode = DispOveride;} // force locked mode
    if (DispNext > 0) {
        switch(DispNext) {
          case 1: Display_Mirrored(); break;
        } DispNext = 0;
    }else {
      if (TEST) {
        // in TEST mode we use a special display
        switch(DispMode) {
          case 0: Display_TestMode(); break;
        }
        DM_Min = 0; DM_Max = 0;
      } else {
        // normal menu screens, display depends on menu modes
        switch(DispMode) {
          case  0: Display_Battery(); break;
          case  1: Display_SenseRaw(); break;
          case  2: Display_SenseAvg(); break;
          case  3: Display_PanTiltErr(); break;
          case  4: Display_SenseAvgAvg(); break;
        }
        DM_Min = 0; DM_Max = 4;
      }
    }
    // the following line puts a rectangle around every display, showing the boundry region
  //  Border = true;
    if (Border) {Display_drawRect(0,0,128,64);}

    // Add a global frame delay offset, if set
    DispDel += DispDelGbl;
  }
}

// --------------------------------------------------------------------------------

void Display_Battery() {
  // normal screen to be displayed
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //               1111111111111111    2222222222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  Display_drawString(64, 14, "Battery");
  if (USB) {
    Display_drawString(64, 34, "USB Mode");
  } else {
    AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
    if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
    if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
    Display_drawString(64, 34, String(float(BatAvg)/BatCal) + "v  " + String(AnyLong) + "%");
  }
  Display_display();
  DispDel = 5;        // wait 5 x 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Battery_Low() {
  // displayed when going into battery low shut-down
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //               1111111111111111    2222222222222222
  Display_setBrightness(25);
  Display_clear();
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  Display_drawString(64,14, "BATTERY");
  Display_drawString(64,34, "LOW");
  Display_display();
  DispDel = 3;        // wait 3 x 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  display.init();
  display.clear();
  display.setContrast(255);
  // display.flipScreenVertically();
  display.display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //       111111111111111111111111    222222222222222222222222
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 6,24,4,24,6
  Display_drawString(64, 6, "Light");
  Display_drawString(64, 34, "Tracker");
  Display_drawRect(0,0,128,64);
  Display_display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_Mirrored() {
  // display robot type when DispMon is first connected
  // we clear current flags so that this is displayed immediately
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 111111111111111111111111 222222222222222222222222   3333333333
  if (Brightness != 255) {Display_setBrightness(255);}  // display may be in sleep mode
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 0,24,2,16,4,10
  Display_drawString(64, 0, "Light");
  Display_drawString(64, 25, "Tracker");
  Display_setFont(10);
  Display_drawString(64,52, "Display Mirror");
  Display_display();
  DispDel = 50;             // wait 50 x 40ms (2s) before displaying next count
}

// --------------------------------------------------------------------------------

void Display_PWM() {
  // display servoAny
  // Display_setFont(24);
  // Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // // Y64 = 6,24,4,24,6
  // int16_t zY = 20;
  // if (WiFiRxRec) {
  //   Display_drawString(64, 6, "WiFi"); zY = 34;
  // }
  // Display_drawString(64, zY, "PWM " + String(ServoAny));
  // DispDisp = true;    // set display flag for next cycle
  // DispDel = 5;        // wait 5 x 40ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_SenseAvg(){
  // display averaged sensor analogue data
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111        2222222222222222        3333333333333333
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, String(Sense0Avg));
  Display_drawString(64,48, String(Sense2Avg));
  Display_drawString(64,24, "Avg.");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 0,24, String(Sense3Avg));
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127,24, String(Sense1Avg));
  Display_display();
  DispDel = 3;        // wait 3 x 40ms (120ms) before displaying next count
}

// --------------------------------------------------------------------------------

void Display_SenseAvgAvg(){
  // display averaged sensor analogue data + average of all 4 sensors
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111        2222222222222222        3333333333333333
  //                             4444444444 
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, String(Sense0Avg));
  Display_drawString(64,48, String(Sense2Avg));
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 0,24, String(Sense3Avg));
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127,24, String(Sense1Avg));
  Display_setFont(10);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64,28, String((Sense0Avg + Sense1Avg + Sense2Avg + Sense3Avg)/4));
  Display_display();
  DispDel = 3;        // wait 3 x 40ms (120ms) before displaying next count
}

// --------------------------------------------------------------------------------

void Display_SenseRaw() {
  // display raw sensor analogue data, compared with calibration points
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  // 1111111111111111        2222222222222222        3333333333333333
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 0, String(Sense0Raw));
  Display_drawString(64,48, String(Sense2Raw));
  Display_drawString(64,24, "Raw");
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString( 0,24, String(Sense3Raw));
  Display_setTextAlignment(TEXT_ALIGN_RIGHT);
  Display_drawString(127,24, String(Sense1Raw));
  Display_display();
  DispDel = 3;        // wait 3 x 40ms (120ms) before displaying next count
}

// --------------------------------------------------------------------------------

void Display_TestMode() {
  // display TEST MODE message
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //       111111111111111111111111    222222222222222222222222
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  Display_drawString(64, 6, "TEST");
  Display_drawString(64, 34, "MODE");
  Display_drawRect(0,0,128,64);
  Display_display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
}

// --------------------------------------------------------------------------------

void DisplayText2(String zT0,String zT1) {
  // display two lines of text of equal height
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //               1111111111111111    2222222222222222
  Display_setFont(16);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  Display_drawString(64, 14, zT0);
  Display_drawString(64, 34, zT1);
  Display_display();  // display this immediately
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------

void DisplayText2L1(String zT0,String zT1) {
  // display two lines of text, 1st line is large
  Display_clear();
  Display_clear();
  // Y64 vertical text spacing, from Y=0 to Y=63
  // 0123456789012345678901234567890123456789012345678901234567890123
  //         111111111111111111111111        2222222222222222
  Display_setFont(24);
  Display_setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 8,24,8,16,8
  Display_drawString(64,  8, zT0);
  Display_setFont(16);
  Display_drawString(64, 40, zT1);
  Display_display();  // display this immediately
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------

void Display_PanTiltErr() {
  // display the Pan and Tilt errors as squares on X,Y axice
  Display_clear();
  Display_setFont(10);
  // calc Tilt error
  int16_t zErr = Sense2Avg - Sense0Avg;
  Display_setTextAlignment(TEXT_ALIGN_LEFT);
  Display_drawString(69, 0, String(zErr));
  Display_drawString(69,11, "Tilt");
  if (zErr < -64) {zErr = -64;}
  if (zErr > 64) {zErr = 64;}
  Display_drawLine(64,0,64,63);
  Display_drawRect(62,32 + (zErr/2) - 2,5,5);
  
  zErr = Sense1Avg - Sense3Avg;
  Display_drawString( 0,36, String(zErr));
  Display_drawString( 0,47, "Pan");
  Display_drawLine( 0,32,127,32);
  if (zErr < -64) {zErr = -64;}
  if (zErr > 64) {zErr = 64;}
  Display_drawRect(64 + zErr - 2,30,5,5);
  Display_display();  // display this immediately
  DispDel = 3;        // wait 3 x 40ms (120ms) before displaying next count
}

// --------------------------------------------------------------------------------
