
// ---------------------------------------------------------------------------------

void attachPanServo() {
  // attaches Pan servo and set its position to PanVal
  if (ESC || USB) return;
  
  servoPanEn = true;
  servoPan.setPeriodHertz(PWM_Freq);    // set PWM frequency
  servoPan.attach(PanPin,500,2400);
  servoPan.writeMicroseconds(PanVal);
  PanAtt = true;
  Serial.println("Pan servo attached.");
}

// ---------------------------------------------------------------------------------

void attachTiltServo() {
  // attaches Tilt servo and set its position to PanVal
  if (ESC || USB) return;
  
  servoTiltEn = true;
  servoTilt.setPeriodHertz(PWM_Freq);    // set PWM frequency
  servoTilt.attach(TiltPin,500,2400);
  servoTilt.writeMicroseconds(TiltVal);
  TiltAtt = true;
  Serial.println("Tilt servo attached.");
}

// --------------------------------------------------------------------------------

void battRead() {
  // read the analogue voltage connected to the battery every 20ms
  BatVol = analogRead(BatPin);
  // BatVol-= BatDist; BatDist++;  // uncomment this line to simulate rapid battery discharge
  BatSum = BatSum + BatVol - BatAvg;
  BatAvg = BatSum/50;     // ADC is averaged over 50 readings to remove noise
  if (USB) {return;}

  // detect critical battery condition and disable HexBot 2
  if (BatAvg <= BatCritical){
    // critical battery voltage reached
    Display_Battery_Low();            // display warning
    // and that's it
    while (true) {yield();}
  }
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // decodes zkeyVal and excutes commands on '.'
  // every time we receive a character set response to 10ms later, just in case
  // there is another packet behind it coming from the controller
  next40ms = millis() - 10;
  if (zkeyVal == 10) {return;}
  if (zkeyVal == 13) {return;}
  keyChar = char(zkeyVal);
  switch (keyChar) {
    case '.': doCmd(); return;          // command terminator
    case '~': // connected to OLED Mirror app via USB or WiFi
      if (!DispMon) {DispMon = true; DispNext = 1;} // display mirrored message
      DispTx = 50; return;
    // case '+': cmdVal = -1; return;      // set flag for ML+. command
    case '-': cmdSgn = -1; return;      // numeric sign
    // case '!': SetMainTask(98); return;  // return to 'Rest' position
    // case '#': Ping = 50; AutoOffTimer = 0;  LEDTaskPnt = 3; return;   // 'ping' sent from application, disables auto-rest function
    case '*': cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1; return; // ESC abort command
    case '0': extendCmdVal(0); return;
    case '1': extendCmdVal(1); return;
    case '2': extendCmdVal(2); return;
    case '3': extendCmdVal(3); return;
    case '4': extendCmdVal(4); return;
    case '5': extendCmdVal(5); return;
    case '6': extendCmdVal(6); return;
    case '7': extendCmdVal(7); return;
    case '8': extendCmdVal(8); return;
    case '9': extendCmdVal(9); return;
  }
  if (cmdMode == ' ') {
    // test for new Command Mode char?
    // only accept certain characters as valid commands, or ignore them
    // delay battery update by 10 seconds whilst receiving commands
    cmdMode = keyChar;
    switch (keyChar) {
      // check for valid mode and convert lower-case
      case 'd': cmdMode = 'D'; break;
      case 'j': cmdMode = 'J'; break;
      case 's': cmdMode = 'S'; break;
    } cmdType = ' '; cmdVal = 0;
  } else {
    // test for Command Type char?
    cmdType = keyChar;
    switch (keyChar) {
      // check for lower-case and convert to upper-case
      case 'a': cmdType = 'A'; break;
      case 'd': cmdType = 'D'; break;
      case 'l': cmdType = 'L'; break;
      case 'r': cmdType = 'R'; break;
      case 't': cmdType = 'T'; break;
      case 'u': cmdType = 'U'; break;
      case 'v': cmdType = 'V'; break;
      case 'x': cmdType = 'X'; break;
      case 'y': cmdType = 'Y'; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void detachPanServo() {
  // detaches Pan servo and set its position to PanVal
  servoPanEn = false;
  servoPan.detach();
  PanTgt = PanVal; // correct target position
  PanAtt = false;
  Serial.println("Pan servo detached.");
}

// ---------------------------------------------------------------------------------

void detachTiltServo() {
  // attaches Tilt servo and set its position to PanVal
  servoTiltEn = false;
  servoTilt.detach();
  TiltTgt = TiltVal; // correct target position
  TiltAtt = false;
  Serial.println("Tilt servo detached.");
}

// ---------------------------------------------------------------------------------

void disableWiFi() {
  // when either a switch is pressed or WiFi disable is required
  WiFiEn = false; WiFiCntC = 0; WiFiRestCnt = 0; C_Dn = 8; Z_Dn = 8;
}

// --------------------------------------------------------------------------------

void displayLft() {
  // change the DispMode pointer to the previous display
  if (DispLock) {return;} // ignore mouse click when display is locked

  DispMode--; if (DispMode < DM_Min) {DispMode = DM_Max;}
  DispDel = 0; DispCnt = 0;
}

// --------------------------------------------------------------------------------

void displayRht() {
  // change the DispMode pointer to the next display
  if (DispLock) {return;} // ignore mouse click when display is locked

  DispMode++; if (DispMode > DM_Max) {DispMode = DM_Min;}
  DispDel = 0; DispCnt = 0;
}

// ---------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received on the serial port so execute command if valid
  // Commands:
  // DDn.     - increase (1) or decrease (-1) DispDel global delay offset DispDelGbl
  // DLn.     - toggle display lock, n=0 OFF, n=1 ON
  // DMxxyy.  - display monitor left mouse click, xx = 0-99, yy = 0-99
  // DM-.     - display monitor mouse up
  // DS.      - display monitor SHIFT-key flag
  // JXnn.    - set global variable JX to nn
  // JYnn.    - set global variable JY to nn
  // SAnn.    - set target servo angle in degrees
  // SD.      - detach the current target servo pin
  // SLnn.    - set target servo lower limit in microseconds
  // SRn.     - set the state of the echo flag, 0 = OFF, 1 = ON
  // STnn.    - set the target servo number 0 - 16, but don't feedback value
  // SUnn.    - set target servo upper limit in microseconds
  // SVnn.    - set target servo angle in microseconds

  cmdVal*= cmdSgn;  // correct cmdVal with the sign value
  
  switch (cmdMode) {
    case ' ': break;
    case 'D':
      switch (cmdType) {
        case 'D': // DDn. command, adjust global delay offset
          if (cmdVal > 0) {DispDelGbl ++;} else {DispDelGbl --;}
          if (DispDelGbl < 0) {DispDelGbl = 0;}
          // display this in the Rx field of Monitor+
          PrintTx+= "$DispDelGbl: " + String(DispDelGbl) +"\n";
          break;
        case 'L':
          if (cmdVal != 0) {DispLock = true; DispOveride = DispMode;}
          else {DispLock = false;}
          break;
        case 'M':
          if (cmdSgn < 0) {DmS = 0;}  // mouse click released
          else {
            if (cmdVal >= 10000) {
              // This is a Display Monitor mouse right-click, DmB = 1
              // determine whether it is a left or right button click
              cmdVal /= 10000; cmdVal --; DmB = 1;
              DmY = cmdVal % 100; DmX = cmdVal/100;
            } else {
              // this is a Display Monitor mouse left-click, DmB = 0
              DmY = cmdVal % 100; DmX = cmdVal/100; DmB = 0;
            }
              // this is not a special 'clickable' display, so simply change it
              if (DmX < 50) {displayLft();  // left side click
              } else {displayRht(); }       // right side click
            DmS = 0;  // cancel SHIFT key flag
          } break;
        case 'S':
          DmS = 1; break; // set SHIFT-key flag
      } break;
    case 'J':
      // commands used to set global variables in association with Joystick app
      switch (cmdType) {
        case 'X': // set global variable JX to cmdVal
          JX = cmdVal; //Serial.println("JX=" + String(JX));
          break;
        case 'Y': // set global variable JY to cmdVal
          JY = cmdVal; //Serial.println("JY=" + String(JY));
          break;
      } break;
    case 'S':
      // keyboard/serial commands used to set servo values
      switch (cmdType) {
        case 'A': // set target servo angle in degrees
          if (ServoPnt > 0) {SetTiltAngle(cmdVal);} 
          else {SetPanAngle(cmdVal);} 
          break;
        case 'D': // detach the current target servo pin
          if (ServoPnt > 0) {TiltAtt = false; servoTilt.detach();}
          else {PanAtt = false; servoPan.detach();}
          break;
        case 'L': // set target servo lower limit in microseconds
          if (ServoPnt > 0) {TiltMin = cmdVal;}
          else {PanMin = cmdVal;}
          break;
        case 'R': // set the state of the Echo flag
          if (cmdVal > 0) {Echo = true; PrintTx += "Echo ON\n";} else {Echo = false;}
          break;
        case 'T': // set target servo pointer in range 0 - 1
          if (cmdVal < 0) {cmdVal = ServoPnt;}
          if (cmdVal > 1) {cmdVal = ServoPnt;}
          ServoPnt = cmdVal; break;
        case 'U': // set target servo upper limit in microseconds
          if (ServoPnt > 0) {TiltMax = cmdVal;}
          else {PanMax = cmdVal;}
          break;
        case 'V': // set target servo angle in microseconds
          if (ServoPnt > 0) {
            // ServoPnt is pointing at Tilt servo
            TiltVal = cmdVal;
            if (!TiltAtt) {attachTiltServo();}
            else {servoTilt.writeMicroseconds(TiltVal);}
          } else {
            // ServoPnt is pointing at Pan servo
            PanVal = cmdVal;
            if (!PanAtt) {attachPanServo();}
            else {servoPan.writeMicroseconds(PanVal);}
          } break;
      } break;
    default:
      PrintTx += "Unknown cmd: " + cmdMode + cmdType + String(cmdVal) + "\n";
      break;
  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
}

// ---------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = abs(cmdVal * 10) + zVal;
}

// --------------------------------------------------------------------------------

int16_t getAdjSense(int16_t zP,int16_t zV) {
  // for a given sensor zP adjust the value zV
  switch(zP) {
    case 0: // sensor 0
           if (zV >= 3561) {zV = map(zV,3561,4053,3490,4000);}
      else if (zV >= 3147) {zV = map(zV,3147,3561,2988,3490);}
      else if (zV >= 2556) {zV = map(zV,2556,3147,2497,2988);}
      else if (zV >= 2048) {zV = map(zV,2048,2556,2000,2497);}
      else if (zV >= 1529) {zV = map(zV,1529,2048,1501,2000);}
      else if (zV >=  995) {zV = map(zV, 995,1529, 999,1501);}
      else if (zV >=  480) {zV = map(zV, 480, 995, 507, 999);}
      else {zV = map(zV,  84, 480,   0, 507);}
      break;
    case 1: // sensor 1
           if (zV >= 3461) {zV = map(zV,3461,3999,3490,4000);}
      else if (zV >= 2952) {zV = map(zV,2952,3461,2988,3490);}
      else if (zV >= 2477) {zV = map(zV,2477,2952,2497,2988);}
      else if (zV >= 1970) {zV = map(zV,1970,2477,2000,2497);}
      else if (zV >= 1483) {zV = map(zV,1483,2048,1501,2000);}
      else if (zV >= 1015) {zV = map(zV,1015,1483, 999,1501);}
      else if (zV >=  532) {zV = map(zV, 532,1015, 507, 999);}
      else {zV = map(zV, 136, 532,   0, 507);}
      break;
    case 2: // sensor 2
           if (zV >= 3444) {zV = map(zV,3444,3932,3490,4000);}
      else if (zV >= 2871) {zV = map(zV,2871,3444,2988,3490);}
      else if (zV >= 2512) {zV = map(zV,2512,2871,2497,2988);}
      else if (zV >= 2070) {zV = map(zV,2070,2512,2000,2497);}
      else if (zV >= 1609) {zV = map(zV,1609,2070,1501,2000);}
      else if (zV >= 1146) {zV = map(zV,1146,1609, 999,1501);}
      else if (zV >=  684) {zV = map(zV, 684,1146, 507, 999);}
      else {zV = map(zV, 314, 684,   0, 507);}
      break;
    case 3: // sensor 3
           if (zV >= 3495) {zV = map(zV,3495,4018,3490,4000);}
      else if (zV >= 2983) {zV = map(zV,2983,3495,2988,3490);}
      else if (zV >= 2445) {zV = map(zV,2445,2983,2497,2988);}
      else if (zV >= 1913) {zV = map(zV,1913,2445,2000,2497);}
      else if (zV >= 1385) {zV = map(zV,1385,1913,1501,2000);}
      else if (zV >=  843) {zV = map(zV, 843,1385, 999,1501);}
      else if (zV >=  333) {zV = map(zV, 333, 843, 507, 999);}
      else {zV = map(zV,   1, 333,   0, 507);}
      break;
  }
  // now constrain the result before returning the value
  if (zV < 0) {zV = 0;}
  if (zV > 4095) {zV = 4095;}
  return zV;
}

// --------------------------------------------------------------------------------

String getBAhex() {
  // Rerturns the current Broardcast Address MAC, ie: 50:02:91:68:F7:3F
  String zRet$ = ""; uint8_t zV;
  for (int zP = 0;zP < 6;zP++) {
    zV = broadcastAddress[zP]/16;
    if (zV < 10) {zRet$ += char(zV + 48);} else {zRet$ += char(zV + 55);}
    zV = broadcastAddress[zP] - (zV *16);
    if (zV < 10) {zRet$ += char(zV + 48);} else {zRet$ += char(zV + 55);}
    if (zP < 5) {zRet$ += ":";}
  }
  return zRet$;
}

// --------------------------------------------------------------------------------

int16_t getSensePcent(int16_t zS) {
  // returns the percentage off sensod zS from the calibrated value
  // vaue returned is in the range -100 to +100
  switch (zS) {
    case 0: // sense0
      if (Sense0Avg < Sense0Ref) {return map(Sense0,0,Sense0Ref,-100,0);}
      else {return map(Sense0,Sense0Ref,4095,0,100);}
      break;
    case 1: // sense1
      if (Sense1Avg < Sense1Ref) {return map(Sense1,0,Sense1Ref,-100,0);}
      else {return map(Sense1,Sense1Ref,4095,0,100);}
      break;
    case 2: // sense2
      if (Sense2Avg < Sense2Ref) {return map(Sense2,0,Sense2Ref,-100,0);}
      else {return map(Sense2,Sense2Ref,4095,0,100);}
      break;
    case 3: // sense3
      if (Sense3Avg < Sense3Ref) {return map(Sense3,0,Sense3Ref,-100,0);}
      else {return map(Sense3,Sense3Ref,4095,0,100);}
      break;
  }
}

// --------------------------------------------------------------------------------

void Init_ESP_NOW() {
  // called to initiate an ESEP-NOW link or connect to another trannsceiver
  // each time this function is called it will attempt to connect to a different
  // boradcast address, cycling through them, over and over
  // Serial.println("Init_ESP_NOW()");

  // if we have previously tried a connection then we need to de-initialise it
  if (ESP_NOW_Init) {
    // already attempted to initialise ESP-NOW so strip everything back
    // Serial.println("Closing ESP-NOW");
    esp_now_del_peer(broadcastAddress);
    // WiFi.disconnect();
  }
  
  // Init ESP-NOW and returns its status
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }
  ESP_NOW_Init = true;
  Serial.println("ESP_NOW_Init = true");
  
  // change the broadcast address pointer each time
  ESP_NOW_BA++;
  // Serial.println("ESP_NOW_BA = " + String(ESP_NOW_BA));
  
  switch (ESP_NOW_BA) {
    case 0: // Wii D1 Mk1 Transciever unique MAC: 50:02:91:68:F7:3F
      broadcastAddress[0] = 0x50;
      broadcastAddress[1] = 0x02;
      broadcastAddress[2] = 0x91;
      broadcastAddress[3] = 0x68;
      broadcastAddress[4] = 0xF7;
      broadcastAddress[5] = 0x3F;
      break;
    case 1: // Wii D1 Mk1 Transciever unique MAC: 50:02:91:68:F0:D2
      broadcastAddress[0] = 0x50;
      broadcastAddress[1] = 0x02;
      broadcastAddress[2] = 0x91;
      broadcastAddress[3] = 0x68;
      broadcastAddress[4] = 0xF0;
      broadcastAddress[5] = 0xD2;
      break;
    case 2: // Wii D1 Mk1 Transciever unique MAC: 58:BF:25:DB:27:A2
      broadcastAddress[0] = 0x58;
      broadcastAddress[1] = 0xBF;
      broadcastAddress[2] = 0x25;
      broadcastAddress[3] = 0xDB;
      broadcastAddress[4] = 0x27;
      broadcastAddress[5] = 0xA2;
      break;
    case 3: // Wii ESP32-C3 Zero Transciever unique MAC: EC:DA:3B:BD:4B:AC
      broadcastAddress[0] = 0xEC;
      broadcastAddress[1] = 0xDA;
      broadcastAddress[2] = 0x3B;
      broadcastAddress[3] = 0xBD;
      broadcastAddress[4] = 0x4B;
      broadcastAddress[5] = 0xAC;
      break;
    default:
      // End of list reached, so extend pause time and reset cycle
      // Turn off serial port reporting after first pass, unless connection fails
      ESP_NOW_BA = -1;      // 1st case will be 0
      WiFiConCnt = 75;      // 3 sec delay between trying cycles
      WiFiTryOnce = false;  // don't report trying after one cycle
      break;
    }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);

  // Register peer
  esp_now_peer_info_t peerInfo; //initialize and assign the peer information as a pointer to an addres
  memset(&peerInfo, 0, sizeof(peerInfo));
  memcpy(peerInfo.peer_addr, broadcastAddress, 6); //copy the value of  broadcastAddress with 6 bytes to peerInfo.peer_addr
  peerInfo.channel = 0;     //channel at which the esp talk. 0 means undefined and data will be sent on the current channel. 1-14 are valid channels which is the same with the local device 
  peerInfo.encrypt = false; //not encrypted

  //Add the device to the paired device list 
  if (esp_now_add_peer(&peerInfo) != ESP_OK){ 
    // Serial.println("Failed to add peer");
    return;
  }

  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
}

// ---------------------------------------------------------------------------------

void PrintAvgs() {
  // called on button press in calibration mode to report values
  Serial.print(String(Sense0Avg) + "," + String(Sense1Avg) + ",");
  Serial.print(String(Sense2Avg) + "," + String(Sense3Avg) + ",");
  Serial.println(String((Sense0Avg + Sense1Avg + Sense2Avg + Sense3Avg)/4));
}

// ---------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set PrintTgt to indicate source as serial port for responses
  keyVal = Serial.read();
  if (keyVal != -1) {SerialRx = true; PrintTgt = 0; decodeKey(keyVal);}
}

// ---------------------------------------------------------------------------------

void readSW0() {
  // called every 10ms
  // read the left button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  sw0State = digitalRead(sw0Pin); // record button state
  if (sw0_Nop) {return;}          // block this function whilst == true
  if (sw0WaitHi && !sw0State) {return;} // wait for sw0 to go HIGH
  sw0WaitHi = false;
  
  bool sw0Clr = false;
  if (sw0State == LOW) {
    //##############################################################################
    //
    // SW0 button is pressed down
    //
    //##############################################################################
    if (sw0LastState == HIGH) {
      // SW0 has just been pressed down
      // Serial.println("SW0 Hi-LO");
      // if (MainTask == 0) {sw0WaitHi = true; SetMainTask(1); return;}  // waking up
      if (Cal) {sw0WaitHi = true; PrintAvgs(); return;}  // send reading to serial monitor
      sw0DwnTime = 0; // restart button down time counter
      sw0Cnt++; // count on the falling edge
      sw0Timer = 0; // reset the timer
    } else {
      // whilst the button is down adjust the timer at 10ms steps
      sw0DwnTime++; // track button pressed time
      if (sw0DwnTime > 150) {
        // SW0 held for 2 seconds so goto rest
        if (MainTask > 0) {sw0Reset(); sw0WaitHi = true; SetMainTask(99);}
      }
    }
  } else {
    //##############################################################################
    //
    // SW0 button is released
    //
    //##############################################################################
    if (sw0LastState == LOW) {
      // SW0 has just been released
      // Serial.println("SW0 Lo-Hi");
    }
    if (sw0Timer >= 50) {
      // button released for 1 sec so assume valid button counts
    //    Serial.print(F("swCnt = ")); Serial.println(swCnt);
        DispMode++; if (DispMode > DM_Max) {DispMode = DM_Min;}
      sw0Cnt = 0; sw0Timer = 0; DispCnt = 0;
    }
    sw0DwnTime = 0;
  }
  sw0LastState = sw0State; // record current state
  if (sw0Cnt > 0) {sw0Timer++;}
}

// ---------------------------------------------------------------------------------

void readSW1() {
  // read the right button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  // SW1 is read every 10ms
  sw1State = digitalRead(sw1Pin); // record button state
  if (sw1_Nop) {return;}          // block this function whilst == true
  if (sw1WaitHi && !sw1State) {return;} // wait for sw0 to go HIGH
  sw1WaitHi = false;
  
  if (sw1State == LOW) {
    //##############################################################################
    //
    // SW1 button is pressed down
    //
    //##############################################################################
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
      // Serial.println("SW1 Hi-LO");
      // if (MainTask == 0) {sw1WaitHi = true; SetMainTask(1); return;}  // waking up
      if (Cal) {sw1WaitHi = true; PrintAvgs(); return;}  // send reading to serial monitor
      sw1DwnTime = 0; // restart button down time counter
      sw1Cnt++; // count on the falling edge
      sw1Timer = 0; // reset the timer
    } else {
      sw1DwnTime++; // track button pressed time
      if (sw1DwnTime > 150) {
        // SW1 held for 2 seconds so goto rest
        if (MainTask > 0) {sw1Reset(); sw1WaitHi = true; SetMainTask(99);}
      }
    }
  } else {
    //##############################################################################
    //
    // sw1State == HIGH as SW1 button is not pressed
    //
    //##############################################################################
    if (sw1LastState == LOW) {
      // SW1 has just been released
      // Serial.println("SW1 Lo-Hi");
    } else {
      if (sw1Timer >= 50) {
        if (sw1Cnt > 0) {
          if (sw1Cnt > 2) {sw1Cnt = 2;}
          SetMainTask(sw1Cnt);
        }
        sw1Cnt = 0; sw1Timer = 0;
      }
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// ---------------------------------------------------------------------------------

void readWiFiRx() {
  // called from the main loop every cycle to check for ESP-NOW data  
  // a state machine is used to separate the data blocks by header type
  // DispDel = 500;  // hold off display updates for 10 sec
  switch(Rx_Task) {
    case 0: // determine the type of data to handle
           if (Rx_Buff.ESPdata[Rx_Pnt] == 'B') {Rx_Task = 3; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'C') {WiiType = 'C'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'N') {WiiType = 'N'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '-') {WiiType = '-'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '$') {Rx_Task = 2; break;}
      // oh dear! bloick header not recognised so discard the block and flag error
      else {WiFiRxErr++; WiFiRx = false;}
      break;
      
    case 1: // load Wii I2C data
      // after the header byte there are 6 data bytes + checksum
      // perform XOR on 6 bytes to check for errors
      if (Rx_Buff.ESPdata[7] == (Rx_Buff.ESPdata[1] ^ Rx_Buff.ESPdata[2] ^ Rx_Buff.ESPdata[3] ^ Rx_Buff.ESPdata[4] ^ Rx_Buff.ESPdata[5] ^ Rx_Buff.ESPdata[6])) {
        // data has passed the checksum test
        RxWiFi[0] = Rx_Buff.ESPdata[1];
        RxWiFi[1] = Rx_Buff.ESPdata[2];
        RxWiFi[2] = Rx_Buff.ESPdata[3];
        RxWiFi[3] = Rx_Buff.ESPdata[4];
        RxWiFi[4] = Rx_Buff.ESPdata[5];
        RxWiFi[5] = Rx_Buff.ESPdata[6];

        // if Nunchuk data strip off the JX,JY and CZ values
        RxRec = true; // indicate a valid frame
        if (WiiType == 'N') {
          JoyX = RxWiFi[0];   // joystick X 0-255
          JoyY = RxWiFi[1];   // joystick Y 0-255
          CZ = RxWiFi[5] & 3; // CZ buttons, C = 2/0, Z = 1/0, active low
        } else {
          // not Nunchuk so set defaults
          JoyX = 128; JoyY = 128; CZ = 3;
        }
      } else {
        WiFiRxErr++;
      }
      Rx_Task = 0;  // reset the task pointer to lok at another block if needed
      Rx_Pnt += 8;  // increase the data pointer beyond this block
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0;} // reached end of buffer
      break;

    case 2: // load text data as if from a serial port
      // read characters up to the end of the buffer
      Rx_Pnt++;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0; break;} // reached end of buffer
      keyVal = Rx_Buff.ESPdata[Rx_Pnt]; PrintTgt = 1; decodeKey(keyVal);
      // Serial.write(keyVal);
      break;

    case 3: // load binary data block
      // data length is specified in second byte
      Rx_Pnt++;
      // at the moment just jump over it, binary transfer is not needed
      Rx_Pnt += Rx_Buff.ESPdata[Rx_Pnt] + 1;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; break;} // reached end of buffer
      break;
  }
}

// --------------------------------------------------------------------------------

void RxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Rx buffer of length zlen
  Rx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Rx_Chk = Rx_Chk ^ Rx_Buff.ESPdata[zP];
  }
}

// --------------------------------------------------------------------------------

void SetMainTask(int zMT) {
  // sets the main task and associated flags
  // immediately stop any movement
  detachPanServo();     // turn OFF PWM signal
  detachTiltServo();    // turn OFF PWM signal
  
  AvgDiv = 32;          // sensor averaging denominator value
  Cal = false;          // == true if in calibration mode
  DispMode = 0;         // default battery display
  ESC = false;
  // MainRun = false;      // clear MainTask inh
  MainTask = zMT;       // set MainTask
  SubCnt = 0;           // sub task counter
  SubTask = 0;          // reset sub task pointer

  switch(MainTask) {
    case  0: break;  // do nothing
    case  1: // wake up and move to vertical position
      break;
    case 99: // go to rest and move to horizontal position
      break;
  }
  // Serial.println("MainTask = " + String(MainTask));
}

// --------------------------------------------------------------------------------

void sw0Reset() {
  // called to reset all of the sw0 variables
  sw0Cnt = 0;             // button switch counter
  sw0DwnTime = 0;         // button pressed down time
  sw0LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw0_Nop = false;        // == true to block SW0 read tasks, default = false
  sw0State = HIGH;        // state of read button switch pin
  sw0Timer = 0;           // timer used to detemine button sequences
  sw0WaitHi = false;      // set == true to wait for a HIGH state
}

// --------------------------------------------------------------------------------

void sw1Reset() {
  // called to reset all of the sw1 variables
  sw1Cnt = 0;             // button switch counter
  sw1DwnTime = 0;         // button pressed down time
  sw1LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw1_Nop = false;        // == true to block SW1 read tasks, default = false
  sw1State = HIGH;        // state of read button switch pin
  sw1Timer = 0;           // timer used to detemine button sequences
  sw1WaitHi = false;      // set == true to wait for a HIGH state
}

// --------------------------------------------------------------------------------

void Wait_ButtonsDwn() {
  // wait in this function until bothe SW0- and SW1 are released
  while (!digitalRead(sw0Pin) || !digitalRead(sw1Pin)) {
    // one or both buttons held down
    yield();  // let system functions run in background
  }
  sw0Cnt = 0; sw0DwnTime = 0; sw0LastState = HIGH; sw0_Nop = false; sw0State = HIGH; sw0Timer = 0;
  sw1Cnt = 0; sw1DwnTime = 0; sw1LastState = HIGH; sw1_Nop = false; sw1State = HIGH; sw1Timer = 0;
  synchLoopTimers();
}

// --------------------------------------------------------------------------------




  //###############################################################################
  //
  //  WiFi Code
  //
  //###############################################################################

void WiFiClearRxBuff(int zNum) {
  // clears the contents of the receive buffer
  for (int zP = 0; zP < zNum;zP++) {Rx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiClearTxBuff(int zNum) {
  // clears the contents of the transmit buffer
  for (int zP = 0; zP < zNum;zP++) {Tx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiDisconnect() {
  // enter the WiFi disconnected state
  WiFiConnected = false;
  WiFiClearRxBuff(WiFiTx_len);
  WiFiClearTxBuff(WiFiTx_len);
  WiFiTryCnt = 0;     // reset the connection retry counter
  WiFiTx_CB = 1;      // set ready to send flag
  WiFiRxRec = false;  // clear the data received flag
  Serial.println("WiFiDisconnected!");
}

// --------------------------------------------------------------------------------

void WiFiTryToConnect() {
  // try to connect with the Wii Transceiver by sending devices name
  // initialise ESP-NOW link first
  Init_ESP_NOW();
  
  Tx_Buff.ESPdata[0] = 'T';
  Tx_Buff.ESPdata[1] = 'r';
  Tx_Buff.ESPdata[2] = 'a';
  Tx_Buff.ESPdata[3] = 'c';
  Tx_Buff.ESPdata[4] = 'k';
  Tx_Buff.ESPdata[5] = 'e';
  Tx_Buff.ESPdata[6] = 'r';
  WiFiTx_len = 7;
  // t0 = micros();
  esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
  if (result == ESP_OK) {
    // Serial.println("Sent with success");
  } else {
    // Serial.println("Error sending the data");
  } 

  WiFiTryNum++; // count the total number of attempts
  // Serial.println("WiFiTryToConnect...");
}

// --------------------------------------------------------------------------------

void WiFiTxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Tx buffer of length zlen
  WiFiTx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    WiFiTx_Chk = WiFiTx_Chk ^ Tx_Buff.ESPdata[zP];
  }
}

// ---------------------------------------------------------------------------------
