
// --------------------------------------------------------------------------------

void PanMoveEngine() {
  // called nominally every 20ms from the main loop to perform Pan movements
  if (PanPause > 0) {
    // reduce PanPause to zero before doing anything
    // PanPause counts in 20ms increments, 50 == 1 second
    if (PanVal == PanTgt) {atPanTgt = true;} else {atPanTgt = false;}
    PanPause--;
  } else {
    // we aren't pausing, so do we need to move the Pan servo
    if (PanTgtCnt > 0) {
      // yes a Pan move is in progress
      if (!PanAtt) {attachPanServo();}
    //  Serial.println(PanTgtCnt);
      atPanTgt = true; // set a test flag for achieving target values
      if (PanTgtCnt == 1) {
        // move Pan servo immediately to the target value
        PanVal = PanTgt; // set values to target values
        servoPan.writeMicroseconds(PanVal);
      } else {
        // move Pan value progressively towards the target value
        servoDelta = PanTgt - PanVal;
        if (abs(servoDelta) > 0) {
          PanVal = PanVal + (servoDelta / PanTgtCnt);
          servoPan.writeMicroseconds(PanVal); atPanTgt = false;
        } if (atPanTgt) {PanTgtCnt = 0;} // already at target value so clear the move count
      }
      PanTgtCnt--; // progressivley reduce the target factor
    //  Serial.println(PanVal);
    }
  }
}

// --------------------------------------------------------------------------------

void MoveToCalibrate() {
  // move to the vertical position for sensor calibration
  // measurements are put into long average mode
  // when a button is pressed we print the sensor values  
  switch (SubTask) {
    case 0: // move to vertical position
      AvgDiv = 500;           // 5 sec sensor averaging denominator value
      SetPanMove(PanCtr); SetTiltMove(TiltCtr);
      DisplayText2("Calibration","Mode");
      Cal = true;             // == true if in calibration mode
      SenseCal = false; SubTask++; break;
    case 1: // wait for vertical to be reached
      if (atPanTgt && atTiltTgt) {DispMode = 99; SubTask++;}
      else {DispDel = 5;}
      break;
  }
}

// --------------------------------------------------------------------------------

void MoveToRest() {
  // user has long pressed a button in active state so move to the rest position
  switch (SubTask) {
    case 0: // move to rest position
      SetPanMove(PanCtr); SetTiltMove(TiltMax);
      DisplayText2("Going to","Rest!");
      SenseCal = false; SubTask++; break;
    case 1: // wait for horizontal to be reached
      if (atPanTgt && atTiltTgt) {DispMode = 0; SetMainTask(0);}
      else {DispDel = 5;} // keep holding off display change
      break;
  }
}

// --------------------------------------------------------------------------------

void MoveToVertical() {
  // user has pressed SW1 once so move to the vertical
  // then go into light tracking mode
  // called every 20ms from main loop
  switch (SubTask) {
    case 0: // move to vertical position
      SetPanMove(PanCtr); SetTiltMove(TiltCtr);
      DisplayText2("Waking","Up!");
      SenseCal = false; SubTask++; break;
    case 1: // wait for vertical to be reached
      if (atPanTgt && atTiltTgt) {SubTask++;}
      else {DispDel = 5;}
      break;
    case 2: // calibrate vertical position
      DisplayText2("Tracking","Mode");
      SubCnt = 100; SubTask++; break;
    case 3: // wait for time-out
      if (SubCnt < 1) {DispMode = 2; SubTask++;}
      SubCnt--; break;
    case 4: // respond to light
      // look at tilt first so that we can take the correct pan action
      // TiltVal increases as we tilt towards the display
      // apply a deadband of 10
    //  Serial.println(String(Sense0) + "," + String(Sense1) + "," + String(Sense2) + "," + String(Sense3));
      ErrTilt = Sense0 - Sense2;
    //  ErrDiv = JX;
      if (ErrTilt > 10) {TiltVal -= (1 + (ErrTilt/ErrDiv));}
      else if (ErrTilt < -10) {TiltVal += (1 - (ErrTilt/ErrDiv));}
      SetTiltPWM(TiltVal);
      
      // PanVal increases as we turn anti-clockwise
      ErrPan = Sense3 - Sense1;
      if (TiltVal < TiltCtr) {
        // looking forwards away from the display
        if (ErrPan > 10) {PanVal += (1 + (ErrPan/ErrDiv));}
        else if (ErrPan < -10) {PanVal -= (1 - (ErrPan/ErrDiv));}
      } else {
        // looking backwards towards the display
        if (ErrPan > 10) {PanVal -= (1 + (ErrPan/ErrDiv));}
        else if (ErrPan < -10) {PanVal += (1 - (ErrPan/ErrDiv));}
      } SetPanPWM(PanVal);
      
      if ((PanVal == PanMax) && (abs(ErrPan) >= Flip)) {
        // trying to go beyond anticlockwise
        if (TiltVal < TiltCtr) {
          // looking forwards away from the display, so flip to
          // looking backwards towards the display
          PanTgt = PanMin + Flip; TiltTgt = TiltCtr + TiltCtr - TiltVal;
          SetPanMove(PanTgt); SetTiltMove(TiltTgt); SubTask = 5;
        } else {
          // looking backwards towards the display, so flip to
          // looking forwards away from the display
          PanTgt = PanMin + Flip; TiltTgt = TiltCtr - TiltVal + TiltCtr;
          SetPanMove(PanTgt); SetTiltMove(TiltTgt); SubTask = 5;
        }
      } else if ((PanVal == PanMin) && (abs(ErrPan) >= Flip)) {
        // trying to go beyond clockwise
        if (TiltVal < TiltCtr) {
          // looking forwards away from the display, so flip to
          // looking backwards towards the display
          PanTgt = PanMax - Flip; TiltTgt = TiltCtr + (TiltCtr - TiltVal);
          SetPanMove(PanTgt); SetTiltMove(TiltTgt); SubTask = 5;
        } else {
          // looking backwards towards the display, so flip to
          // looking forwards away from the display
          PanTgt = PanMax - Flip; TiltTgt = TiltCtr - TiltVal + TiltCtr;
          SetPanMove(PanTgt); SetTiltMove(TiltTgt); SubTask = 5;
        }
      }
      if (SubTask == 5) {PanTgtCnt = 16; TiltTgtCnt = 16;}
      break;
    case 5: // wait for the flip over to be reached
      if (atPanTgt && atTiltTgt) {SubCnt = 5; SubTask++;}
      break;
    case 6: // wait 50ms for new sensor readings to take place
      SubCnt--; if (SubCnt < 1) {SubTask = 4;}
      break;
  }
}

// --------------------------------------------------------------------------------

void SetPanAngle(int zA) {
  // sets Pan servo to an angle zA immediately
  // the mapping function is split either side of 90 for accuracy
  // PanMin is fully clockwise
  // PanMax is fully anticlockwise
  if (ESC) return;
  
  if (zA >= 90) {PanVal = map(zA,90,180,PanCtr,PanMax);}
  else {PanVal = map(zA,0,90,PanMin,PanCtr);}
  PanTgt = PanVal; servoPan.writeMicroseconds(PanVal);
  if (!PanAtt) {attachPanServo();}
}

// --------------------------------------------------------------------------------

void SetPanMove(int16_t zSV) {
  // initiate a Pan servo move to target zSV
  if (zSV == PanVal) {atPanTgt = true; return;}
  
  if (zSV > PanMax) {zSV = PanMax;}
  if (zSV < PanMin) {zSV = PanMin;}
  PanTgt = zSV; PanTgtCnt = 1 + (abs(PanTgt - PanVal)/16);
  atPanTgt = false;
}

// --------------------------------------------------------------------------------

void SetPanPWM(int16_t zPWM) {
  // sets the PWM value of the Pan servo immediately
  // checks value against min/max
  if (zPWM > PanMaxDef) {zPWM = PanMaxDef;}  
  else if (zPWM < PanMinDef) {zPWM = PanMinDef;}
  PanVal = zPWM; servoPan.writeMicroseconds(PanVal);
  if (!PanAtt) {attachPanServo();}
}

// --------------------------------------------------------------------------------

void SetTiltAngle(int zA) {
  // sets Tilt servo to an angle zA immediately
  // the mapping function is split either side of 90 for accuracy
  if (ESC) return;
  
  if (zA >= 90) {TiltVal = map(zA,90,180,TiltCtr,TiltMax);}
  else {TiltVal = map(zA,0,90,TiltMin,TiltCtr);}
  TiltTgt = TiltVal; servoTilt.writeMicroseconds(TiltVal);
  if (!TiltAtt) {attachTiltServo();}
}

// --------------------------------------------------------------------------------

void SetTiltMove(int16_t zSV) {
  // initiate a Tilt servo move to target zSV
  if (zSV == TiltVal) {atTiltTgt = true; return;}
  
  if (zSV > TiltMax) {zSV = TiltMax;}
  if (zSV < TiltMin) {zSV = TiltMin;}
  TiltTgt = zSV; TiltTgtCnt = abs(TiltTgt - TiltVal)/16;
  atTiltTgt = false;
}

// --------------------------------------------------------------------------------

void SetTiltPWM(int16_t zPWM) {
  // sets the PWM value of the Tilt servo immediately
  // checks value against min/max
  if (zPWM > TiltMaxDef) {zPWM = TiltMaxDef;}  
  else if (zPWM < TiltMinDef) {zPWM = TiltMinDef;}
  TiltVal = zPWM; servoTilt.writeMicroseconds(TiltVal);
  if (!TiltAtt) {attachTiltServo();}
}

// --------------------------------------------------------------------------------

void TiltMoveEngine() {
  // called nominally every 20ms from the main loop to perform Tilt movements
  if (TiltPause > 0) {
    // reduce TiltPause to zero before doing anything
    // TiltPause counts in 20ms increments, 50 == 1 second
    if (TiltVal == TiltTgt) {atTiltTgt = true;} else {atTiltTgt = false;}
    TiltPause--;
  } else {
    // we aren't pausing, so do we need to move the Tilt servo
    if (TiltTgtCnt > 0) {
      // yes a Tilt move is in progress
      if (!TiltAtt) {attachTiltServo();}
    //  Serial.println(TiltTgtCnt);
      atTiltTgt = true; // set a test flag for achieving target values
      if (TiltTgtCnt == 1) {
        // move Tilt servo immediately to the target value
        TiltVal = TiltTgt; // set values to target values
        servoTilt.writeMicroseconds(TiltVal);
      } else {
        // move Tilt value progressively towards the target value
        servoDelta = TiltTgt - TiltVal;
        if (abs(servoDelta) > 0) {
          TiltVal = TiltVal + (servoDelta / TiltTgtCnt);
          servoTilt.writeMicroseconds(TiltVal); atTiltTgt = false;
        } if (atTiltTgt) {TiltTgtCnt = 0;} // already at target value so clear the move count
      }
      TiltTgtCnt--; // progressivley reduce the target factor
    //  Serial.println(TiltVal);
    }
  }
}


// --------------------------------------------------------------------------------
